/*  wmbday
 *  Copyright (C) 2003 astrAtis <steffen@x-berg.de>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public Licensse as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 */


#include "config.h"
#include "main.h"


/* global variables */
char* enum_months[] = { "January", "February", "March", "April", "May", "June", "July", "August", "September", "October", "November", "December" };
Display *display;
Window iconwin;
GC theGC;
struct ring* nextbdays[4];
int todays_bday = 0;
int day = 0; /* current day */
struct settings mySettings; /* holds global program settings */


void get_user_homedir() {
	struct passwd *user;
	
	if(!(user = getpwuid(geteuid()))) {
		fprintf(stderr, "Couldn't determine user's homedir\n");
		exit(1);
		return;
	}
	
	if(snprintf(mySettings.datafile, sizeof(mySettings.datafile), "%s/.wmbday/data", user->pw_dir) > sizeof(mySettings.datafile)) {
		fprintf(stderr, "Path to datafile exceeded buffer size of %d chars\n", sizeof(mySettings.datafile));
		exit(1);
	}
	return;
}


void ParseCMDLine(int argc, char **argv) {
	int c;	
#ifdef LINUX
	struct option const long_options[] =
	{
		{"bgcolor", required_argument, 0, 'B'},
		{"fontcolor", required_argument, 0, 'F'},
		{"hicolor", required_argument, 0, 'H'},
		{"file", required_argument, 0, 'f'},
		{"help", no_argument, 0, 'h'},
		{"version", no_argument, 0, 'v'},
		{NULL, 0, NULL, 0}
	};
#endif
	struct stat statbuf;

	/* set default values first */
	mySettings.color_bg = GetColor("white");
	mySettings.color_font = GetColor("black");
	mySettings.color_hilight = GetColor("red");
	
#ifdef LINUX
	while((c = getopt_long(argc, argv, "f:B:F:H:hv", long_options, NULL)) != -1) {
#endif
#ifdef FREEBSD
	while((c = getopt(argc, argv, "f:B:F:H:hv")) != -1) {
#endif
		switch(c) {
			case 'B': /* bgcolor */
				if(optarg)
					mySettings.color_bg = GetColor(optarg);
				else {
					fprintf(stderr, "no bg-color given\n");
					exit(1);
				}
				break;
			case 'F': /* fontcolor */
				if(optarg)
					mySettings.color_font = GetColor(optarg);
				else {
					fprintf(stderr, "no font-color given\n");
					exit(1);
				}
				break;
			case 'H': /* hicolor */
				if(optarg)
					mySettings.color_hilight = GetColor(optarg);
				else {
					fprintf(stderr, "no highlight-color given\n");
					exit(1);
				}
				break;
			case 'f': /* data-file */
				if(optarg) {
					if(snprintf(mySettings.datafile, sizeof(mySettings.datafile), "%s", optarg) > sizeof(mySettings.datafile)) {
						fprintf(stderr, "Path to datafile exceeded buffer size of %d chars\n", sizeof(mySettings.datafile));
						exit(1);
					}
					
					/* check if it is a file */
					if(stat(mySettings.datafile, &statbuf) != 0) {
						perror("stat");
						exit(1);
					}
					
					if(!S_ISREG(statbuf.st_mode)) {
						fprintf(stderr, "Data file specified is not a regular file.\n");
						exit(1);
					}					
				}
				else {
					printf("no file given\n");
					exit(1);
				}
				break;
			case 'h': /* help */
				PrintHelp();
				break;
			case 'v': /* version */
				PrintVersion();
				break;
			
			default:
				PrintHelp();
				break;
		}
	}		
	return;
}


void PrintHelp() {
	printf("Usage: wmbday [options]\n");
	printf("  -f, --file <file>        use <file> as data file instead of ~/.wmbday/data\n");
	printf("  -B, --bgcolor <color>    sets color (as text) used for filling the background\n");
	printf("  -F, --fontcolor <color>  sets color (as text) used for text\n");
	printf("  -H, --hicolor <color>    sets color (as text) used for text when blinking\n");
	printf("  -h, --help               outputs this text..\n");
	printf("  -v, --version            prints version\n");
	exit(0);
	return;
}


void PrintVersion() {
	printf("wmbday " VERSION "\n");
	printf("Copyright (C) 2003 astrAtis <steffen@x-berg.de>\n");
	exit(0);
	return;
}


Window CreateWindow() {
	XClassHint	classHint;
	XWMHints	thewmhints;
	Pixmap		pixmask;
	Window		window;
	
	/* connect to x server */
	if((display = XOpenDisplay(NULL)) == NULL) {
		fprintf(stderr, "Couldn't open display\n");
		return 1;
	}
	
	/* create the window */
	window = XCreateSimpleWindow(display, RootWindow(display, 0), 0, 0, 64, 64, 1, BlackPixel(display, DefaultScreen(display)), WhitePixel(display, DefaultScreen(display)));
	iconwin = XCreateSimpleWindow(display, window, 0, 0, 64, 64, 1, BlackPixel(display, DefaultScreen(display)), WhitePixel(display, DefaultScreen(display)));

	/* put pixmask over window so that outer border becomes transparent */
	pixmask = XCreateBitmapFromData(display, window, wmbday_bits, wmbday_width, wmbday_height);
	XShapeCombineMask(display, window, ShapeBounding, 0, 0, pixmask, ShapeSet);
	XShapeCombineMask(display, iconwin, ShapeBounding, 0, 0, pixmask, ShapeSet);
	
	/* set class hint */
	classHint.res_name = "wmbday";
	classHint.res_class = "wmbday";
	XSetClassHint(display, window, &classHint);

	/* make it a dockapp */
	thewmhints.initial_state = WithdrawnState;
	thewmhints.window_group = window;
	thewmhints.icon_window = iconwin;
	thewmhints.icon_x = 0;
	thewmhints.icon_y = 0;
	thewmhints.flags = StateHint | WindowGroupHint | IconWindowHint | IconPositionHint;
	XSetWMHints(display, window, &thewmhints);
	
	/* require event-notification */
	XSelectInput(display, window, ExposureMask | PointerMotionMask | ButtonPressMask);
	XSelectInput(display, iconwin, ExposureMask | PointerMotionMask | ButtonPressMask);
	
	return window;
}


void signal_handler(int sig) {
	time_t tnow;
	struct tm *now;
	
	static int run = 0;
	int i;
	
	switch(sig) {
		case SIGCHLD: /* avoid zombie */
			waitpid(-1, NULL, WNOHANG);
			break;
		
		case SIGALRM: /* animate the text */
			/* draw text */
			XSetForeground(display, theGC, run & 1 ? mySettings.color_font.pixel : mySettings.color_hilight.pixel);
	
			for(i = 0; i < todays_bday; i++)
				XDrawString(display, iconwin, theGC, 8, 15+i*14, nextbdays[i]->name, strlen(nextbdays[i]->name));
			
			run = !run;	
			break;
			
		case SIGHUP: /* reload data file */
			fprintf(stderr, "Caught SIGHUP. Reloading datafile...\n");

			/* load datafile, get next birthdays again and redraw window so that changes are being shown */
			LoadDatafile();
			
			tnow = time(NULL);
			now = localtime(&tnow);
			ring_goto_date(now->tm_mday, now->tm_mon+1);
			ring_get_next(nextbdays, 4);
			
			day = 0; /* let main-loop check for today's birthdays again */
			
			RedrawWindow(NULL);
			break;
	}
}


int LoadDatafile(void) {
	FILE *fp;
	char buffer[16];
	int count;
	
	/* if there is a ring, we clear it here first */
	ring_clear();
	
	/* open data file */
	if((fp = fopen(mySettings.datafile, "r")) == NULL) {
		fprintf(stderr, "Couldn't open datafile: %s\n", mySettings.datafile);
		exit(1);
		return 1;
	}

	/* check data file format (simple or vcard) and call appropriate parser */
	fgets(buffer, 12, fp);
	rewind(fp);
	
	if (!strcmp(buffer, "BEGIN:VCARD")) 
		count = _LoadDatafile_vCard(fp);
	else 
		count = _LoadDatafile_Simple(fp);

	/* don't start, if no birthdays could be loaded */
	if(count == 0) {
		printf("No birthdays could be loaded, exiting.\n");
		exit(0);
	}
	else
		printf("Loaded %d birthdays\n", count);	
	
	/*close file and free memory */
	fclose(fp);
	
	return 0;
}


int _LoadDatafile_Simple(FILE *fp) {
	char buffer[128];
	char* buf;
	int day, month, year;
	int count = 0;
	int line = 0;
	char* ptrchar;
	
	while (fgets(buffer, 101, fp)) {
		++line;
		
		buf = buffer;
		
		/* ignore whitespace at the beginning of line and lines that are comments */
		buf += strspn(buf, WHITESPACE);
		if(*buf == '#')
			continue;
		
		/* first character of line has to be a number */		
		if(!isdigit(*buf)) {
			fprintf(stderr, "data file: line %d does not begin with a number\n", line);
			continue;
		}
		
		/* read the date */
		sscanf(buf, "%2d.%2d.%4d", &day, &month, &year);
		
		/* check date */
		if(is_date(day, month, year) == 0) {
			fprintf(stderr, "data file: invalid date in line %d. (Out of range or in future)\n", line);
			continue;
		}
		
		/* jump behind the date */
		buf += strcspn(buf, WHITESPACE);
		buf += strspn(buf, WHITESPACE);
		
		/* handle names in "" */
		if(*buf == '\"') {
			++buf;
			ptrchar = buf + strcspn(buf, "\"");
			*(ptrchar) = '\0';
		}
		else {
			ptrchar = buf + strcspn(buf, "\n\r");
			*ptrchar = '\0';
		}
		
		/* check if there is a name in this line */
		if(!strlen(buf)) {
			fprintf(stderr, "data file: no name given in line %d.\n", line);
			continue;
		}
		
		ring_add(buf, day, month, year);
		
		++count;
	}
	
	return count;
}


int _LoadDatafile_vCard(FILE *fp) {
	char buffer[128];
	char* ptr;
	int day, month, year;
	char name[96];
	int count = 0;

	int name_found = 0, bday_found = 0;

	while (fgets(buffer, 100, fp)) {
		if (!strncmp(buffer, "FN:", 3)) {
			strncpy(name, buffer+3, sizeof(name));
			ptr = strrchr(name, '\n');
			*ptr = '\0';
			name_found = 1; continue;
		}
		if (!strncmp(buffer, "BDAY:", 5)) {
			sscanf(buffer+5, "%4d-%2d-%2d", &year, &month, &day);
			bday_found = 1; continue;
		}
		if (!strncmp(buffer, "END:VCARD", 9)) {
			if (name_found && bday_found) {
				ring_add(name, day, month, year);
				++count;
			}
			name_found = bday_found = 0;
		}
	}

	return count;
}


int CreateWindowGC() {
	XGCValues values;
	
	if((theGC = XCreateGC(display, RootWindow(display, DefaultScreen(display)), 0, &values)) == 0) {
		fprintf(stderr, "Couldn't create window GC\n");
		return 1;
	}
	return 0;
}


void RedrawWindow(XEvent* theevent) {
	int i;
	
	XSetForeground(display, theGC, mySettings.color_bg.pixel);
	if(theevent != NULL)
		XFillRectangle(display, iconwin, theGC, theevent->xexpose.x, theevent->xexpose.y, theevent->xexpose.width, theevent->xexpose.height);
	else
		XFillRectangle(display, iconwin, theGC, 0, 0, 64, 64);
	
	XSetForeground(display, theGC, mySettings.color_font.pixel);
	
	for(i=0; i<4; i++) {
		if(nextbdays[i] != NULL)
			XDrawString(display, iconwin, theGC, 8, 15+i*14, nextbdays[i]->name, strlen(nextbdays[i]->name));
	}
	return;
}


XColor GetColor(char* name) {
	XColor color;
	Colormap colormap;

	colormap = DefaultColormap(display, DefaultScreen(display));

    if(!XParseColor(display, colormap, name, &color))
		fprintf(stderr, "Couldn't parse \"%s\"\n", name);
    if(!XAllocColor(display, colormap, &color))
		fprintf(stderr, "Couldn't alloc \"%s\"\n", name);
	
    return color;
}


void StartTimer() {
	struct itimerval timer;
		
	timer.it_value.tv_sec = 0;
	timer.it_value.tv_usec = 250000;
	
	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = 250000;
	
	setitimer(ITIMER_REAL, &timer, NULL);
	
	return;
}


void StopTimer() {
	struct itimerval timer;

	timer.it_value.tv_sec = 0;
	timer.it_value.tv_usec = 0;

	timer.it_interval.tv_sec = 0;
	timer.it_interval.tv_usec = 0;
	
	setitimer(ITIMER_REAL, &timer, NULL);
	RedrawWindow(NULL);
	return;
}


int main(int argc, char **argv) {
	XEvent event;
	time_t tnow;
	struct tm *now;
	Colormap colormap;
	Window window;
	struct itimerval timerinfo;
	int row;
	int days_left;
	int rollover = 0;
	int fork_child;
	char buffer[1024];

	mySettings.datafile[0] = 0x0;
	
	/* create the main window */	
	if((window = CreateWindow()) == 0)
		return 0;
	
	/* parse parameters from command line */
	ParseCMDLine(argc, argv);
	
	/* if no datafile was specified at command line, we set to default */
	if(mySettings.datafile[0] == 0x0)
		get_user_homedir();
	
	/* now show the window */
	XMapWindow(display, window);
	XFlush(display);
	
	/* create a graphics context for window for drawing */
	CreateWindowGC();

	/* set handler on SIGHUP which makes wmbday reload the data from data-file */
	if(signal(SIGHUP, signal_handler) == SIG_ERR) {
		fprintf(stderr, "Couldn't set signal handler for SIGHUP\n");
		return 1;
	}
	
	/* set handler on SIGALRM for animation-timer */
	if(signal(SIGALRM, signal_handler) == SIG_ERR) {
		fprintf(stderr, "Couldn't set signal handler for SIGALRM\n");
		return 1;
	}
	
	/* set handler on SIGCHLD to avoid zombies */
	if(signal(SIGCHLD, signal_handler) == SIG_ERR) {
		fprintf(stderr, "Couldn't set signal handler for SIGCHLD\n");
		return 1;
	}
	
	/* load data from data file into ring */
	LoadDatafile();
	
	/* main loop */
	while(1) {
		tnow = time(NULL);
    	now = localtime(&tnow);
		
		/* day changed */
		if(day != now->tm_mday){
			ring_goto_date(now->tm_mday, now->tm_mon+1);
			ring_get_next(nextbdays, 4);
			RedrawWindow(NULL);
			
			if((todays_bday = ring_check_bday(now->tm_mday, now->tm_mon+1)) > 0)
				StartTimer();
			
			day = now->tm_mday;
		}
		
		while(XPending(display)) {
			XNextEvent(display, &event);
			switch(event.type) {
				case Expose: /* redraw window */
					RedrawWindow(&event);
					break;
				
				case ButtonPress: /* mouse click */
					/* left mouse button */	
					if(event.xbutton.button == 1) {
						getitimer(ITIMER_REAL, &timerinfo);
						
						/* show info about clicked person */
						if(timerinfo.it_value.tv_usec == 0) {
							row = event.xbutton.y / 16;
							if(nextbdays[row] != NULL) {
								days_left = days_between_dates(day, now->tm_mon + 1, nextbdays[row]->day, nextbdays[row]->month, now->tm_year + 1900);
								
								/* for people whose bday is next year we have to add 1 when calculating their age */
								if(compare_dates(day, now->tm_mon + 1, nextbdays[row]->day, nextbdays[row]->month) == 1)
										rollover = 1;
								
								switch(days_left) {
									case 0:
										snprintf(buffer, sizeof(buffer), "%s\nhas a birthday today! The lucky person turned %d.", nextbdays[row]->name, now->tm_year+1900 - nextbdays[row]->year);
										break;
									case 1:
										snprintf(buffer, sizeof(buffer), "%s\nhas a birthday in %d day on %s %d\nand will turn %d then.", nextbdays[row]->name, days_left, enum_months[nextbdays[row]->month-1], nextbdays[row]->day, now->tm_year+1900 - nextbdays[row]->year + rollover);
										break;
									default:
										snprintf(buffer, sizeof(buffer), "%s\nhas a birthday in %d days on %s %d\nand will turn %d then.", nextbdays[row]->name, days_left, enum_months[nextbdays[row]->month-1], nextbdays[row]->day, now->tm_year+1900 - nextbdays[row]->year + rollover);
										break;
								}
								
								/* fork() to create a messagebox with xmessage */
								if((fork_child = fork()) < 0) {
									fprintf(stderr, "Couldn't fork(): %s\n", strerror(errno));
									exit(1);
								}
								
								if(fork_child == 0) {
									if(execlp(XM_PATH, "xmessage", "-center", buffer, NULL) < 0) {
										fprintf(stderr, "Couldn't execute \"xmessage\": %s\n", strerror(errno));
										exit(1);
									}
									exit(0);
								}
							}
						}
						else 
							StopTimer(); /* stop animation */
					}
					break;
			}
		}
		usleep(10000);
	}
	
	/* free colors */
	colormap = DefaultColormap(display, DefaultScreen(display));
	XFreeColors(display, colormap, &mySettings.color_font.pixel, 1, 0);
	XFreeColors(display, colormap, &mySettings.color_hilight.pixel, 1, 0);
	XFreeColors(display, colormap, &mySettings.color_bg.pixel, 1, 0);
	
	/* close connection to xserver */
	XCloseDisplay(display);
	return 0;
}
